import logging

from dbnd.testing.helpers_pytest import assert_run_task
from dbnd_examples.orchestration.customizations.custom_target import CustomIOPipeline


logger = logging.getLogger(__name__)


class TestRunWithCustomTargets(object):
    def test_parameter_with_multiple_partitions(self):
        assert_run_task(CustomIOPipeline())



import logging

from typing import Dict, List

import matplotlib
import pandas as pd

from matplotlib.figure import Figure
from pandas import DataFrame, np

from dbnd import band, log_dataframe, output, parameter, task
from dbnd.tasks.basics.publish import publish_results
from dbnd.utils import data_combine
from targets import Target


matplotlib.use("Agg")
logger = logging.getLogger(__name__)


@task
def dump_table_from_db(table, db_name="master"):
    # type: (str, str) -> List[str]
    # output: raw_id(table+id), partner, value
    return ["raw_id,partner,value\n"] + [
        "%s_%s,%s,%s\n" % (table, idx, idx % 3, idx * 2) for idx in range(10)
    ]


@band
def dump_db(tables, db_name="master"):
    # type: (List[str], str) -> Dict[str, pd.DataFrame]
    result = {
        t: dump_table_from_db(table=t, db_name=db_name, task_name="table_%s" % t)
        for t in tables
    }
    return result


@task
def filter_partner(data, partner):
    # type: (DataFrame, int) -> DataFrame
    partner_data = data[data["partner"] == partner]
    log_dataframe("partner_%s" % partner, partner_data)
    return partner_data


@task(result=output.folder[Dict[str, DataFrame]])
def separate_data_by_type(data):
    # type: (DataFrame) -> Dict[str, DataFrame]

    return {"a": data.head(1), "b": data.head(2), "c": data.head(1)}


@task
def build_report(partners_files, partners_df):
    # type: (Dict[int,Target], DataFrame) -> str
    logger.info("partners files: %s", partners_files)
    logger.info("partners data: %s", partners_df)
    return "OK"


@task
def build_graphs(partner_info):
    # type: (DataFrame) -> Figure
    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax1 = fig.add_subplot(2, 2, 1)
    ax1.hist(np.random.randn(100), bins=20, alpha=0.3)
    ax2 = fig.add_subplot(2, 2, 2)
    ax2.scatter(np.arange(30), np.arange(30) + 3 * np.random.randn(30))
    fig.add_subplot(2, 2, 3)
    return fig


@band
def partners_report(partners):
    # type: (List[int]) -> ...
    table = "aa"

    dump = dump_db(tables=[table])
    partners_info = {
        partner: filter_partner(dump[table], partner) for partner in partners
    }
    by_type = separate_data_by_type(dump[table])

    report = build_report(
        partners_files=partners_info, partners_df=data_combine(partners_info.values())
    )

    graphs = {partner: build_graphs(pi) for partner, pi in partners_info.items()}

    published = publish_results(graphs).published

    return report, by_type, published



####rewrote

from pandas import DataFrame

from dbnd import dbnd_tracking
from dbnd_examples.data import data_repo


class TestDocPythonScripts:
    def test_doc(self):
        #### DOC START
        import logging
        import os

        from typing import Tuple

        from dbnd import dbnd_tracking, task

        @task
        def prepare_data(data: DataFrame = data_repo.wines):
            prepared_data = data
            logging.info(prepared_data)
            return data

        @task
        def join_data(base_data: DataFrame, extra_data: DataFrame):
            return base_data.join(extra_data)

        @task
        def join_data_pipeline(dataframes_number=3):
            data = prepare_data(data_repo.wines)
            for i in range(dataframes_number):
                data = join_data(data, prepare_data(data_repo.wines))
            return data

        @task
        def prepare_all_data(dataframes_number=3) -> Tuple[DataFrame, DataFrame]:
            data = DataFrame.empty
            for i in range(dataframes_number):
                data = prepare_data(data_repo.wines)

            join_data_pipe = join_data_pipeline()
            return data, join_data_pipe

        # if __name__ == "__main__":
        # os.environ["DBND__CORE__DATABAND_URL"] = "<url>"
        # os.environ["DBND__CORE__DATABAND_ACCESS_TOKEN"] = "<access_token>"
        with dbnd_tracking():
            prepare_all_data()
        #### DOC END



import mock

from dbnd_redshift.dbnd_redshift import (
    get_redshift_query_id,
    log_redshift_resource_usage,
)


def redshift_snippet(redshift_conn):
    query_id = get_redshift_query_id(
        "select * from my_table limit 20", connection_string=redshift_conn
    )
    log_redshift_resource_usage(query_id, connection_string=redshift_conn)


class TestDbTracking(object):
    def test_redshift_code(self):
        with mock.patch("psycopg2.connect") as redshift_conn:
            redshift_snippet(redshift_conn)



from datetime import timedelta

from airflow import DAG
from airflow.operators.python_operator import PythonOperator
from airflow.utils.dates import days_ago

from dbnd_airflow.scheduler.dagrun_zombies import find_and_kill_dagrun_zombies


default_args = {
    "owner": "airflow",
    "start_date": days_ago(1),
    "execution_timeout": timedelta(minutes=5),
}

with DAG(
    dag_id="dbnd_clear_zombies",
    default_args=default_args,
    schedule_interval=timedelta(minutes=5),
    description="Maintenance DAG to remove staled task (zombies)",
) as dag:
    PythonOperator(
        task_id="clear_zombies",
        python_callable=find_and_kill_dagrun_zombies,
        op_args=(None,),
    )



import datetime
import logging

from typing import List, Tuple

import pyspark.sql as spark

from pyspark.ml import Pipeline
from pyspark.ml.classification import LogisticRegression
from pyspark.ml.feature import VectorAssembler
from pyspark.mllib.evaluation import RegressionMetrics
from pyspark.sql.functions import exp
from pyspark.sql.types import DoubleType

from dbnd import log_metric, parameter, pipeline
from dbnd.utils import data_combine, period_dates
from dbnd_examples.data import demo_data_repo
from dbnd_spark.spark import spark_task
from targets import target
from targets.types import PathStr


logger = logging.getLogger(__name__)


def data_source(task_target_date, name):
    return target(demo_data_repo.partner_to_file(name, task_target_date))


@spark_task(result=parameter.output.csv[spark.DataFrame])
def get_and_enrich_spark(raw_data: spark.DataFrame, column_name: str):
    raw_data.show()
    data_with_new_feature = raw_data.withColumn(column_name + "_exp", exp(column_name))
    return data_with_new_feature


@spark_task(result=parameter.output.csv[spark.DataFrame])
def clean_data_spark(raw_data: spark.DataFrame):
    return raw_data.na.fill(0)


@pipeline()
def ingest_partner_a(task_target_date):
    raw_data = data_source(name="a", task_target_date=task_target_date)
    clean = clean_data_spark(raw_data=raw_data)
    return get_and_enrich_spark(raw_data=clean, column_name="1")


@pipeline
def ingest_partner_b(task_target_date):
    raw_data = data_source(name="b", task_target_date=task_target_date)
    return clean_data_spark(raw_data=raw_data)


@pipeline
def ingest_partner_c(task_target_date):
    raw_data = data_source(name="c", task_target_date=task_target_date)
    clean = clean_data_spark(raw_data=raw_data)
    return get_and_enrich_spark(raw_data=clean, column_name="10")


@pipeline
def fetch_partner_data_spark(
    task_target_date, selected_partners: List[str], period=datetime.timedelta(days=7)
) -> List[spark.DataFrame]:
    partner_data = []
    for partner in selected_partners:
        all_data = []
        for d in period_dates(task_target_date, period):
            if partner == "a":
                data = ingest_partner_a(task_target_date=d)
            elif partner == "b":
                data = ingest_partner_b(task_target_date=d)
            elif partner == "c":
                data = ingest_partner_c(task_target_date=d)
            else:
                raise Exception("Partner not found!")

            all_data.append(data)
        partner_data.append(data_combine(all_data, sort=True))
    return partner_data


@spark_task
def calculate_features(
    raw_data: List[spark.DataFrame], selected_features: List[str] = None
) -> spark.DataFrame:
    result = raw_data.pop(0)
    for d in raw_data:
        result = result.join(d, ["id"], "outer")
    if selected_features:
        result = result.select(selected_features)
    result.show()
    return result


@spark_task(result="training_set, test_set, validation_set")
def split_data_spark(
    raw_data: spark.DataFrame,
) -> Tuple[spark.DataFrame, spark.DataFrame, spark.DataFrame]:

    columns_to_remove = set(["id", "0_norm", "10_norm"])
    if columns_to_remove.issubset(list(raw_data.schema.names)):
        raw_data = raw_data.drop(columns_to_remove)

    (train, test) = raw_data.randomSplit([0.8, 0.2])
    (test, validation) = raw_data.randomSplit([0.5, 0.5])

    target_stats = raw_data.describe(["target"])

    log_metric(
        "target.mean",
        target_stats.filter(target_stats["summary"] == "mean")
        .collect()[0]
        .asDict()["target"],
    )
    log_metric(
        "target.std",
        target_stats.filter(target_stats["summary"] == "stddev")
        .collect()[0]
        .asDict()["target"],
    )

    return train, test, validation


@spark_task
def train_model_spark(
    test_set: spark.DataFrame,
    training_set: spark.DataFrame,
    alpha: float = 1.0,
    l1_ratio: float = 0.5,
    saved_model=parameter.output.folder_data.with_flag(None)[PathStr],
) -> str:

    transform = VectorAssembler(inputCols=["0", "1", "2"], outputCol="features")
    lr = LogisticRegression(
        featuresCol="features",
        labelCol="target",
        regParam=l1_ratio,
        elasticNetParam=alpha,
        family="multinomial",
        maxIter=1,
    )
    ppl = Pipeline(stages=[transform, lr])

    # Fit the pipeline to training documents.
    model = ppl.fit(training_set)

    prediction = model.transform(test_set)
    evaluation = prediction.withColumn(
        "label", prediction["target"].cast(DoubleType())
    ).select(["label", "prediction"])
    evaluation.show()
    metrics = RegressionMetrics(evaluation.rdd)

    log_metric("r2", metrics.r2)
    log_metric("alpha", alpha)

    model.write().save(str(saved_model))
    return "ok"


@pipeline(result=("model"))
def train_model_for_customer_spark(
    task_target_date,
    data: spark.DataFrame = None,
    alpha: float = 1.0,
    l1_ratio: float = 0.5,
    period=datetime.timedelta(days=1),
    selected_features: List[str] = None,
):
    if data is None:
        partners = fetch_partner_data_spark(
            task_target_date=task_target_date, period=period
        )
        data = calculate_features(
            selected_features=selected_features, raw_data=partners
        )

    training_set, test_set, validation_set = split_data_spark(raw_data=data)

    model = train_model_spark(
        test_set=test_set, training_set=training_set, alpha=alpha, l1_ratio=l1_ratio
    )

    return model.saved_model



from datetime import timedelta

import airflow

from airflow import DAG
from airflow.operators.python_operator import PythonOperator

from dbnd import dont_track
from dbnd_airflow import track_dag


class TestDocTrackingAirflowDags:
    def test_prepare_data_operator(self):
        #### DOC START
        from airflow.models import BaseOperator

        from dbnd import log_metric, task

        class PrepareData(BaseOperator):
            def execute(self, context):
                calculate_alpha()
                prepare_data()

        def calculate_alpha():
            log_metric("alpha", 0.5)

        @task
        def prepare_data():
            log_metric("lines", 10)
            log_metric("chars", 300)

        #### DOC END
        test_preparedata = PrepareData(task_id="test_PrepareData").execute("<CONTEXT>")

    def test_dont_track(self):
        #### DOC START
        @dont_track
        def calculate_beta():
            return 1.5

        #### DOC END
        calculate_beta()

    def test_track_specific_dags(self):
        #### DOC START
        args = {"start_date": airflow.utils.dates.days_ago(2)}
        dag = DAG(
            dag_id="calculate_alpha",
            default_args=args,
            schedule_interval=timedelta(days=1),
        )

        def calculate_alpha():
            return 0.5

        PythonOperator(
            task_id="calculate_alpha", python_callable=calculate_alpha, dag=dag
        )

        track_dag(dag)
        #### DOC END



import logging

from dbnd.testing.helpers_pytest import assert_run_task
from dbnd_examples.orchestration.examples.salad import prepare_salad


logger = logging.getLogger(__name__)


class TestTutorialSyntax(object):
    def test_prepare_salad(self):
        assert_run_task(prepare_salad.task())



from dbnd_examples.data import data_repo


class TestDocPysparkDataframe:
    def test_doc(self):
        #### DOC START
        import sys

        from operator import add

        from pyspark.sql import SparkSession

        from dbnd import log_metric, task

        @task
        def prepare_data(data, output_file):
            spark = SparkSession.builder.appName("PythonWordCount").getOrCreate()

            lines = spark.read.text(data).rdd.map(lambda r: r[0])
            counts = (
                lines.flatMap(lambda x: x.split(" "))
                .map(lambda x: (x, 1))
                .reduceByKey(add)
            )
            counts.saveAsTextFile(output_file)
            output = counts.collect()
            for (word, count) in output:
                print("%s: %i" % (word, count))
            log_metric("counts", len(output))
            spark.sparkContext.stop()

        if __name__ == "__main__":
            if len(sys.argv) != 3:
                print("Usage: wordcount  ")
                sys.exit(-1)
            prepare_data(sys.argv[1], sys.argv[2])
        #### DOC END
        # prepare_data.dbnd_run(data_repo.vegetables, "~/Data/example")



'''
class TestDocTrackingSnowflakeGuide:
    def test_doc(self):
        #### DOC START
        import snowflake.connector
        from dbnd_snowflake import log_snowflake_resource_usage, log_snowflake_table

        def establish_connection():
            """Establish and return a Snowflake Connection using Snowflake Python Connector"""
            snowflake_conn_id = "snowflake_connection"
            snowflake_connection = snowflake.connector.connect(
                user=your_username,
                password=your_password,
                account=your_full_account_name,
            )
            return snowflake_connection

        def run_query(connection, query):
            """Generate a cursor, and execute a query. Returns the query result, query ID and session ID."""
            cursor = connection.cursor()
            query_results = cursor.execute(query).fetchall()
            query_id, session_id = cursor.sfqid, cursor.connection.session_id
            cursor.close()
            return query_results, query_id, session_id

        def query_snowflake():
            """Query Snowflake, track resource usage, and log Snowflake table"""
            connection_string = f"snowflake://your_connection_string"
            snowflake_connection = establish_connection()
            query_ids = []

            database = "SNOWFLAKE_SAMPLE_DATA"
            schema = "TPCDS_SF100TCL"
            table_name = "STORE_SALES"
            query = f'select * from "{database}"."{schema}"."{table_name}" limit 30'

            results, query_id, session_id = run_query(snowflake_connection, query)
            query_ids.append(query_id)

            log_snowflake_resource_usage(
                database=database,
                connection_string=connection_string,
                query_ids=query_ids,
                session_id=int(session_id),
                delay=1,
            )

            log_snowflake_table(
                table_name=f"{table_name}",
                connection_string=connection_string,
                database=database,
                schema=schema,
                with_preview=False,
            )

            # As best practice, close the Snowflake connection
            snowflake_connection.close()

        if __name__ == "__main__":
            query_snowflake()
        #### DOC END
        query_snowflake()
'''



class TestDocTrackingDatabricks:
    def test_doc(self):
        #### DOC START
        from random import randint
        from time import sleep

        import pandas as pd

        from dbnd import dataset_op_logger, dbnd_tracking, log_metric

        def execute():
            for int in range(randint(0, 10)):
                sleep(randint(0, 10))
                log_metric(f"interation_{int}", int)

            with dataset_op_logger("databricks://test/load/read", "read") as logger:
                data = {"row_1": [3, 2, 1, 0], "row_2": ["a", "b", "c", "d"]}
                read_df = pd.DataFrame.from_dict(data, orient="index")
                logger.set(data=read_df)

            with dataset_op_logger("databricks://test/load/write", "write") as logger:
                data = {"row_1": [3, 2, 1, 0], "row_2": ["a", "b", "c", "d"]}
                write_df = pd.DataFrame.from_dict(
                    data, orient="index", columns=["A", "B", "C", "D"]
                )
                logger.set(data=write_df)

        with dbnd_tracking("databricks_test"):
            execute()
        #### DOC END



import datetime
import logging

from pathlib import Path
from typing import List

import six

from dbnd import task
from targets import Target
from targets.types import PathStr


logger = logging.getLogger(__name__)


@task
def f_task_parameters(
    v_int: int,
    v_bool: bool,
    v_str: str,
    v_datetime: datetime.datetime,
    v_date: datetime.date,
    v_period: datetime.timedelta,
    v_list: List,
    v_list_obj: List[object],
    v_list_str: List[str],
    v_list_int: List[int],
    v_list_date: List[datetime.date],
) -> object:
    result = (
        v_int,
        v_datetime,
        v_str,
        v_date,
        v_period,
        v_list,
        v_list_obj,
        v_list_str,
        v_list_int,
        v_list_date,
    )
    logger.info("v_int: %s", v_int)
    logger.info("v_bool: %s", v_bool)
    logger.info("v_datetime: %s", v_datetime)
    logger.info("v_list: %s", v_list)
    logger.info("v_list_date: %s", v_list_date)

    return result


@task
def f_path_types(
    pathlib_path: Path, str_as_path: PathStr, target_path: Target
) -> object:
    assert isinstance(pathlib_path, Path)
    assert isinstance(str_as_path, six.string_types)
    assert isinstance(target_path, Target)

    return (pathlib_path, str_as_path, target_path)



from typing import List

from dbnd.testing.helpers_pytest import assert_run_task
from dbnd_examples.orchestration.examples.top_artists import (
    top_artists_big_report,
    top_artists_report,
)


class TestTopArtistExamples(object):
    def test_top_artists_report(self):
        task = assert_run_task(top_artists_report.task())
        actual = task.result.load(List[str])
        assert len(actual) == 10

    def test_top_artists_big_report(self):
        task = assert_run_task(top_artists_big_report.task())
        actual = task.result.load(List[str])
        assert len(actual) == 15



"""from random import random

class TestDocTrackingMlflow:
    def test_doc(self):
        #### DOC START
        from dbnd import task
        from mlflow import start_run, end_run
        from mlflow import log_metric, log_param

        @task
        def calculate_alpha():
            start_run()
            # params
            log_param("alpha", random())
            log_param("beta", random())
            # metrics
            log_metric("accuracy", random())
            log_metric("coefficient", random())
            end_run()

        #### DOC END
        calculate_alpha.dbnd_run()"""



# rewrote


class TestDocMetrics:
    def test_doc(self):
        #### DOC START
        from dbnd import log_metric

        def calculate_alpha(alpha, beta):
            sum = alpha + beta
            log_metric("sum", sum)

            return alpha

        calculate_alpha(0.5, 0.03)
        #### DOC END



from os import path

from setuptools import setup
from setuptools.config import read_configuration


BASE_PATH = path.join(path.dirname(path.abspath(__file__)))
CFG_PATH = path.join(BASE_PATH, "setup.cfg")
config = read_configuration(CFG_PATH)

version = config["metadata"]["version"]

EXTRAS_REQUIRE = {"airflow": ["dbnd-airflow[airflow]==" + version]}

INSTALL_REQUIRES = [
    # we are still installing 'databand' in dockers.. "dbnd==" + version,
    "dbnd-airflow==" + version,
    "scikit-learn==0.23.2",
    "scipy==1.1.0",
    "matplotlib==3.3.0",
    "pandas<2.0.0,>=0.17.1",
]

setup(
    name="dbnd-examples",
    package_dir={"": "src"},
    version=version,
    zip_safe=False,
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    entry_points={"dbnd": ["dbnd-examples = dbnd_examples._plugin"]},
)



from pandas import DataFrame


#### DOC START
# module1.py
def calculate_alpha():
    pass


def calculate_beta():
    pass


def calculate_gamma():
    pass


#### DOC END


class TestDocFunctions:
    def test_tracking_functions_with_decorator(self):
        #### DOC START
        # module1.py
        from dbnd import task

        # define a function with a decorator

        @task
        def prepare_data(data: DataFrame, counter: int, random: int):
            return "OK"

        #### DOC END

    def test_tracking_named_functions(self):
        #### DOC START
        # module2.py

        from dbnd import track_functions

        # from module1 import calculate_alpha, calculate_beta, calculate_gamma

        track_functions(calculate_alpha, calculate_beta, calculate_gamma)

        def calculate_coefficient():
            calculate_alpha()
            calculate_beta()
            calculate_gamma()

        # module3.py

        """from dbnd import track_module_functions
        import module1
        import module2
        from module2 import calculate_coefficient

        track_modules(module1, module2)"""

        def calculate_all():
            calculate_coefficient()

        #### DOC END
        calculate_all()

    def test_alternate_calculate_coefficient(self):
        #### DOC START
        # module2.py

        from dbnd import track_module_functions

        """from module1 import calculate_alpha, calculate_beta, calculate_gamma
        import module1

        track_module_functions(module1)"""

        def calculate_coefficient():
            calculate_alpha()
            calculate_beta()
            calculate_gamma()

        #### DOC START
        calculate_coefficient()
